"""Data interface functions for the mudpy engine."""

# Copyright (c) 2004-2017 Jeremy Stanley <fungi@yuggoth.org>. Permission
# to use, copy, modify, and distribute this software is granted under
# terms provided in the LICENSE file distributed with this software.

import os
import re
import stat

import mudpy
import yaml


class DataFile:

    """A file containing universe elements and their facets."""

    def __init__(self, filename, universe):
        self.filename = filename
        self.universe = universe
        self.data = {}
        self.load()

    def load(self):
        """Read a file, create elements and poplulate facets accordingly."""
        self.modified = False
        try:
            self.data = yaml.safe_load(open(self.filename))
        except FileNotFoundError:
            # it's normal if the file is one which doesn't exist yet
            log_entry = ("File %s is unavailable." % self.filename, 6)
            try:
                mudpy.misc.log(*log_entry)
            except NameError:
                # happens when we're not far enough along in the init process
                self.universe.setup_loglines.append(log_entry)
        if not hasattr(self.universe, "files"):
            self.universe.files = {}
        self.universe.files[self.filename] = self
        includes = []
        if "__control__" in self.data:
            if "include_files" in self.data["__control__"]:
                for included in self.data["__control__"]["include_files"]:
                    included = find_file(
                        included,
                        relative=self.filename,
                        universe=self.universe)
                    if included not in includes:
                        includes.append(included)
            if "include_dirs" in self.data["__control__"]:
                for included in [
                    os.path.join(x, "__init__.yaml") for x in
                        self.data["__control__"]["include_dirs"]
                ]:
                    included = find_file(
                        included,
                        relative=self.filename,
                        universe=self.universe
                    )
                    if included not in includes:
                        includes.append(included)
            if "default_files" in self.data["__control__"]:
                origins = self.data["__control__"]["default_files"]
                for key in origins.keys():
                    origins[key] = find_file(
                        origins[key],
                        relative=self.filename,
                        universe=self.universe
                    )
                    if origins[key] not in includes:
                        includes.append(origins[key])
                    self.universe.default_origins[key] = origins[key]
                    if key not in self.universe.categories:
                        self.universe.categories[key] = {}
            if "private_files" in self.data["__control__"]:
                for item in self.data["__control__"]["private_files"]:
                    item = find_file(
                        item,
                        relative=self.filename,
                        universe=self.universe
                    )
                    if item not in includes:
                        includes.append(item)
                    if item not in self.universe.private_files:
                        self.universe.private_files.append(item)
        for node in list(self.data):
            if node == "__control__":
                continue
            facet_pos = node.rfind(".") + 1
            if not facet_pos:
                mudpy.misc.Element(node, self.universe, self.filename,
                                   old_style=True)
            else:
                prefix = node[:facet_pos].strip(".")
                try:
                    element = self.universe.contents[prefix]
                except KeyError:
                    element = mudpy.misc.Element(prefix, self.universe,
                                                 self.filename)
                element.set(node[facet_pos:], self.data[node])
                if prefix.startswith("mudpy.movement."):
                    self.universe.directions.add(
                        prefix[prefix.rfind(".") + 1:])
        for include_file in includes:
            if not os.path.isabs(include_file):
                include_file = find_file(
                    include_file,
                    relative=self.filename,
                    universe=self.universe
                )
            if (include_file not in self.universe.files or not
                    self.universe.files[include_file].is_writeable()):
                DataFile(include_file, self.universe)

    def save(self):
        """Write the data, if necessary."""
        normal_umask = 0o0022
        private_umask = 0o0077
        private_file_mode = 0o0600

        # when modified, writeable and has content or the file exists
        if self.modified and self.is_writeable() and (
           self.data or os.path.exists(self.filename)
           ):

            # make parent directories if necessary
            if not os.path.exists(os.path.dirname(self.filename)):
                old_umask = os.umask(normal_umask)
                os.makedirs(os.path.dirname(self.filename))
                os.umask(old_umask)

            # backup the file
            if "__control__" in self.data and "backup_count" in self.data[
                    "__control__"]:
                max_count = self.data["__control__"]["backup_count"]
            elif "mudpy.limit" in self.universe.contents:
                max_count = self.universe.contents["mudpy.limit"].get(
                    "backups", 0)
            else:
                max_count = 0
            if os.path.exists(self.filename) and max_count:
                backups = []
                for candidate in os.listdir(os.path.dirname(self.filename)):
                    if re.match(
                       os.path.basename(self.filename) +
                       r"""\.\d+$""", candidate
                       ):
                        backups.append(int(candidate.split(".")[-1]))
                backups.sort()
                backups.reverse()
                for old_backup in backups:
                    if old_backup >= max_count - 1:
                        os.remove(self.filename + "." + str(old_backup))
                    elif not os.path.exists(
                        self.filename + "." + str(old_backup + 1)
                    ):
                        os.rename(
                            self.filename + "." + str(old_backup),
                            self.filename + "." + str(old_backup + 1)
                        )
                if not os.path.exists(self.filename + ".0"):
                    os.rename(self.filename, self.filename + ".0")

            # our data file
            if self.filename in self.universe.private_files:
                old_umask = os.umask(private_umask)
                file_descriptor = open(self.filename, "w")
                if oct(stat.S_IMODE(os.stat(
                        self.filename)[stat.ST_MODE])) != private_file_mode:
                    # if it's marked private, chmod it appropriately
                    os.chmod(self.filename, private_file_mode)
            else:
                old_umask = os.umask(normal_umask)
                file_descriptor = open(self.filename, "w")
            os.umask(old_umask)

            # write and close the file
            yaml.safe_dump(self.data, allow_unicode=True,
                           default_flow_style=False, stream=file_descriptor)
            file_descriptor.close()

            # unset the modified flag
            self.modified = False

    def is_writeable(self):
        """Returns True if the __control__ read_only is False."""
        try:
            return not self.data["__control__"].get("read_only", False)
        except KeyError:
            return True


def find_file(
    file_name=None,
    prefix=None,
    relative=None,
    search=None,
    stash=None,
    universe=None
):
    """Return an absolute file path based on configuration."""

    # this is all unnecessary if it's already absolute
    if file_name and os.path.isabs(file_name):
        return os.path.realpath(file_name)

    # if a universe was provided, try to get some defaults from there
    if universe:

        if hasattr(
                universe, "contents") and "mudpy.filing" in universe.contents:
            filing = universe.contents["mudpy.filing"]
            if not prefix:
                prefix = filing.get("prefix")
            if not search:
                search = filing.get("search")
            if not stash:
                stash = filing.get("stash")

        # if there's only one file loaded, try to work around a chicken<egg
        elif hasattr(universe, "files") and len(
            universe.files
        ) == 1 and not universe.files[
                list(universe.files.keys())[0]].is_writeable():
            data_file = universe.files[list(universe.files.keys())[0]].data

            # try for a fallback default directory
            if not stash:
                stash = data_file.get(".mudpy.filing.stash", "")

            # try for a fallback root path
            if not prefix:
                prefix = data_file.get(".mudpy.filing.prefix", "")

            # try for a fallback search path
            if not search:
                search = data_file.get(".mudpy.filing.search", "")

        # another fallback root path, this time from the universe startdir
        if hasattr(universe, "startdir"):
            if not prefix:
                prefix = universe.startdir
            elif not os.path.isabs(prefix):
                prefix = os.path.join(universe.startdir, prefix)

    # when no root path is specified, assume the current working directory
    if not prefix:
        prefix = os.getcwd()

    # make sure it's absolute
    prefix = os.path.realpath(prefix)

    # if there's no search path, just use the root path and etc
    if not search:
        search = [prefix, "etc"]

    # work on a copy of the search path, to avoid modifying the caller's
    else:
        search = search[:]

    # if there's no default path, use the last component of the search path
    if not stash:
        stash = search[-1]

    # if an existing file or directory reference was supplied, prepend it
    if relative:
        if os.path.isdir(relative):
            search = [relative] + search
        else:
            search = [os.path.dirname(relative)] + search

    # make the search path entries absolute and throw away any dupes
    clean_search = []
    for each_path in search:
        if not os.path.isabs(each_path):
            each_path = os.path.realpath(os.path.join(prefix, each_path))
        if each_path not in clean_search:
            clean_search.append(each_path)

    # start hunting for the file now
    for each_path in clean_search:

        # if the file exists and is readable, we're done
        if os.path.isfile(os.path.join(each_path, file_name)):
            file_name = os.path.realpath(os.path.join(each_path, file_name))
            break

    # it didn't exist after all, so use the default path instead
    if not os.path.isabs(file_name):
        file_name = os.path.join(stash, file_name)
    if not os.path.isabs(file_name):
        file_name = os.path.join(prefix, file_name)

    # and normalize it last thing before returning
    file_name = os.path.realpath(file_name)

    # normalize the resulting file path and hand it back
    return file_name
